// mml2nes

#include <stdio.h>
#include <malloc.h>
#include <memory.h>
#include <ctype.h>

#define CH_MAX         4
#define EVENT_MAX   4096
#define LABEL_MAX   1024
#define LENTBL_SIZE   15
#define LENTBL_MAX   256
#define TONE_RANGE    15
#define ENV_MAX       15
#define ENV_SIZE     128
#define DT_SIZE       15
#define RP_NAME_SIZE  64
#define RP_MAX        32

#define BASE_TICK 633600 // 2*2*2*2*2*2*2*2*3*3*5*5*11

#define TYPE_UNKNOWN    0
#define TYPE_NOTE       1
#define TYPE_REST       2
#define TYPE_REG1       3
#define TYPE_REG2       4
#define TYPE_TONE_BASE  5
#define TYPE_LEN_TABLE  6
#define TYPE_ENV_BASE   7
#define TYPE_TIE        8
#define TYPE_PERCUSS    9
#define TYPE_REPEAT    10
#define TYPE_JMP       11
#define TYPE_CALL      12
#define TYPE_RET       13
#define TYPE_END       14
#define TYPE_LABEL     16
#define TYPE_SKIP      17

#define DEF_EP 0x00004550
#define DEF_DT 0x00004454
#define DEF_RP 0x00005250

//           value        len        quant
// NOTE                      quantl
// REST      0                   quantl
// REG1      ݒl       0          0
// REG2      ݒl       0          0
// TONE_BASE ()   () 0
// LEN_TABLE LTableԍ   0          0
// ENV_BASE  envԍ      0          0
// TIE       0            0          quantl
// PERCUSS   0            0          0
// REPEAT    JԂ xԍ [(0,1)
// JMP       0            xԍ short/long
// CALL      QƔԍ     xԍ 0
// RET       0            0          0
// END       0            0          0
// LABEL     QƔԍ     xԍ 0
// SKIP      0            xԍ 0
//
// xԍ gbN[v:0 ̑1ȏ

typedef struct {
	unsigned short next;
	unsigned char  type;
	unsigned char  quant;
	unsigned short value;
	unsigned short len;
} event_t;

typedef struct {
	unsigned short head;
	unsigned short last;
	unsigned short last_note;
	int base_tone;
	int base_len;
	int def_len;
	int quant;
	int label;
	int save_base_tone[2];
	int save_base_len[2];
	int save_def_len[2];
	int save_quant[2];
	int exist_skip[2];
	int assigned_label[2];
	int nest;
	int ref_ev;
} seq_t;

typedef struct {
	unsigned short len[LENTBL_SIZE];
	unsigned char  quant[LENTBL_SIZE];
	int count;
	int base;
} lentbl_t;

typedef struct {
	unsigned char data[ENV_SIZE];
	int count;
	int base;
} env_t;

typedef struct {
	unsigned short data[DT_SIZE];
	int count;
} dt_t;

typedef struct {
	char name[RP_NAME_SIZE];
	unsigned short data;
} rp_t;

typedef struct {
	event_t  events[EVENT_MAX];
	lentbl_t lentbl[LENTBL_MAX];
	env_t env[ENV_MAX];
	rp_t rhythm[RP_MAX];
	seq_t ch[CH_MAX];
	dt_t dt;
	int evcnt;
	int ltcnt;
	int rpcnt;
} score_t;

typedef struct {
	int pos;
	int count;
} call_t;

typedef struct {
	int pos;
	int low;
	int high;
	int ltbl;
} flow_t;

typedef struct {
	flow_t near;
	flow_t far;
	int link;
} opttbl_t;

int parse(score_t *score, FILE *fp);
int insert_loop_event(score_t *score);
int insert_label_event(score_t *score);
int insert_tone_event(score_t *score);
int insert_lentbl_event(score_t *score);
int optimize_event(score_t *score);
int assign_label_offset(score_t *score);
int assign_envtable_base(score_t *score);
int write_cmd(score_t *score, FILE *fp, const char *prefix);
int write_data(score_t *score, FILE *fp, const char *prefix);
void dump(score_t *score);

int main(int argc, char *argv[]) {
	FILE *ifp = NULL;
	FILE *cfp = NULL;
	FILE *dfp = NULL;
	const char *mml_name;
	const char *cmd_name = "out_cmd.asm";
	const char *dat_name = "out_data.asm";
	const char *prefix = "Out";
	score_t *score;
	int i;

	if (argc < 2) {
		printf("Usage: %s <mml file> [<cmd source file>] [<data source file>] [<prefix>]\n", argv[0]);
	}

	score = (score_t *)malloc(sizeof(score_t));
	if (score == NULL) goto fail;
	memset(score, 0, sizeof(score_t));

	mml_name = argv[1];
	if (argc >= 3) {
		cmd_name = argv[2];
	}
	if (argc >= 4) {
		dat_name = argv[3];
	}
	if (argc >= 5) {
		prefix = argv[4];
	}

	ifp = fopen(mml_name, "r");
	if (ifp == NULL) goto fail;

	cfp = fopen(cmd_name, "w");
	if (cfp == NULL) goto fail;

	dfp = fopen(dat_name, "w");
	if (dfp == NULL) goto fail;

	parse(score, ifp);
	insert_loop_event(score);
	insert_label_event(score);
	insert_tone_event(score);
	insert_lentbl_event(score);
	optimize_event(score);
	assign_label_offset(score);
	assign_envtable_base(score);
	write_cmd(score, cfp, prefix);
	write_data(score, dfp, prefix);
//	dump(score);

fail:
	if (ifp) fclose(ifp);
	if (cfp) fclose(cfp);
	if (dfp) fclose(dfp);
	free(score);

	return 0;
}

void append_event(score_t *score, int i, int type, int value, int len, int quant) {
	event_t *ev;

	score->evcnt++;
	ev = score->events + score->evcnt;

	if (score->ch[i].head == 0) {
		score->ch[i].head = score->evcnt;
		score->ch[i].last = score->evcnt;
	} else {
		score->events[score->ch[i].last].next = score->evcnt;
		score->ch[i].last = score->evcnt;
	}

	ev->next  = 0;
	ev->type  = type;
	ev->value = value;
	ev->len   = len;
	ev->quant = quant;
}

void insert_event(score_t *score, int i, int prev, int type, int value, int len, int quant) {
	event_t *ev;

	score->evcnt++;
	ev = score->events + score->evcnt;

	if (score->ch[i].last == prev) {
		score->ch[i].last = score->evcnt;
	}
	if (prev == 0) {
		ev->next = score->ch[i].head;
		score->ch[i].head = score->evcnt;
	} else {
		ev->next = score->events[prev].next;
		score->events[prev].next = score->evcnt;
	}

	ev->type  = type;
	ev->value = value;
	ev->len   = len;
	ev->quant = quant;
}

char *parse_int(char *p, int *result, int num) {
	int value = 0;

	if (*p == '$') {
		p++;
		while (isxdigit(*p)) {
			if (isdigit(*p)) {
				value = value * 16 + (*p - '0');
			} else if (isupper(*p)) {
				value = value * 16 + (*p - 'A' + 10);
			} else {
				value = value * 16 + (*p - 'a' + 10);
			}
			p++;
		}
	} else {
		while (isdigit(*p)) {
			value = value * 10 + (*p - '0');
			p++;
		}
	}

	*result = value;
	return p;
}

char *parse_len(char *p, int *result, int base_len, int def_len, int num) {
	char err_buf[64];
	char *q;
	int len, len2, len3;
	int i;

	q = p;
	len3 = 0;
	for (;;) {
		if (isdigit(*p) || (*p == '$')) {
			p = parse_int(p, &len, num);
			if (base_len % len) goto err_len;
			len = base_len / len;
		} else {
			len = base_len / def_len;
		}
		len2 = len;
		while (*p == '.') {
			if (len2 % 2) goto err_len;
			len2 /= 2;
			len += len2;
			p++;
		}
		len3 += len;

		if (*p != '^') break;
		p++;
	}

	*result = len3;
	return p;

err_len:
	p = q;
	q = err_buf;
	for (;;) {
		if (*p == '$') {
			*(q++) = *(p++);
			while (isxdigit(*p)) *(q++) = *(p++);
		} else {
			while (isdigit(*p)) *(q++) = *(p++);
		}
		while (*p == '.')  *(q++) = *(p++);
		if (*p != '^') break;
		p++;
	}
	*q = '\0';
	printf("%d: invalid tone length (%s)\n", num, err_buf);
	return p;
}

int parse_define(score_t *score, int def, char *p, int num) {
	char buf[4];
	char *q;
	int index;
	int value;

	if (def == DEF_EP) {
		while (isspace(*p)) p++;
		p = parse_int(p, &index, num);
		if ((index >= 1) && (index <= ENV_MAX)) {
			while (isspace(*p)) p++;
			if (*p == ',') p++;
			while (isspace(*p)) p++;
			score->env[index-1].count = 0;
			while (isdigit(*p) || (*p == '$')) {
				p = parse_int(p, &value, num);
				if (score->env[index-1].count < ENV_SIZE) {
					score->env[index-1].data[score->env[index-1].count++] = value & 0xff;
				}
				while (isspace(*p)) p++;
				if (*p == ',') p++;
				while (isspace(*p)) p++;
			}
		} else {
			printf("%d: invalid index of EP.\n", num);
		}
	} else if (def == DEF_DT) {
		while (isspace(*p)) p++;
		if (*p == ',') p++;
		while (isspace(*p)) p++;
		score->dt.count = 0;
		while (isdigit(*p) || (*p == '$')) {
			p = parse_int(p, &value, num);
			if (score->dt.count < DT_SIZE) {
				score->dt.data[score->dt.count++] = value & 0xffff;
			}
			while (isspace(*p)) p++;
			if (*p == ',') p++;
			while (isspace(*p)) p++;
		}
	} else if (def == DEF_RP) {
		if (score->rpcnt < RP_MAX) {
			index = score->rpcnt++;
			score->rhythm[index].data = 0;

			while (isspace(*p)) p++;
			if (*p == ',') p++;
			while (isspace(*p)) p++;
			q = score->rhythm[index].name;
			while (isalpha(*p) || (*p == '_')) {
				*(q++) = toupper(*p++);
			}
			*q = '\0';

			while (isspace(*p)) p++;
			if (*p == ',') p++;
			while (isspace(*p)) p++;
			p = parse_int(p, &value, num);
			score->rhythm[index].data |= ((value & 7) << 12);

			while (isspace(*p)) p++;
			if (*p == ',') p++;
			while (isspace(*p)) p++;
			p = parse_int(p, &value, num);
			score->rhythm[index].data |= ((value & 0x8f) << 8);

			while (isspace(*p)) p++;
			if (*p == ',') p++;
			while (isspace(*p)) p++;
			p = parse_int(p, &value, num);
			score->rhythm[index].data |= ((value & 15) << 4);

			while (isspace(*p)) p++;
			if (*p == ',') p++;
			while (isspace(*p)) p++;
			p = parse_int(p, &value, num);
			score->rhythm[index].data |= (value & 15);
		} else {
			printf("%d: Rhythm define is over.\n", num);
		}
	} else {
		// `
		q = buf;
		if (def & 0x00ff0000) *(q++) = (def>>16) & 0xff;
		if (def & 0x0000ff00) *(q++) = (def>> 8) & 0xff;
		if (def & 0x000000ff) *(q++) = (def    ) & 0xff;
		*q = '\0';
		printf("%d: syntax error. (#%s)\n", num, q);
	}

	return 0;
}

int parse_score(score_t *score, int ch, char *p, int num) {
	static int tone_inx[] = { 9, 11, 0, 2, 4, 5, 7 };
	char name[RP_NAME_SIZE];
	event_t *ev;
	char *q;
	int value;
	int len;
	int nest;
	int i;

	while (*p) {
		if (isspace(*p)) {
			p++;
		} else if (*p == ';') {
			// Rg
			while (*p) p++;
		} else if (*p == '@') {
			p++;
			if (*p == 'x') {
				// WX^1w
				p++;
				if (isdigit(*p) || (*p == '$')) {
					p = parse_int(p, &value, num);
					if ((value >= 0) && (value < 256)) {
						for (i = 0; i < CH_MAX; i++) {
							if (ch & (1 << i)) {
								append_event(score, i, TYPE_REG1, value, 0, 0);
							}
						}
					} else {
						printf("%d: invalid value in reg1.\n", num);
					}
				} else {
					printf("%d: no value in reg1.\n", num);
				}
			} else if (*p == 'y') {
				// WX^2w
				p++;
				if (isdigit(*p) || (*p == '$')) {
					p = parse_int(p, &value, num);
					if ((value >= 0) && (value < 256)) {
						for (i = 0; i < CH_MAX; i++) {
							if (ch & (1 << i)) {
								append_event(score, i, TYPE_REG2, value, 0, 0);
							}
						}
					} else {
						printf("%d: invalid value in reg2.\n", num);
					}
				} else {
					printf("%d: no value in reg2.\n", num);
				}
			} else if ((*p == 'e') && (*(p+1) == 'v')) {
				// Gx[vw
				p += 2;
				if (isdigit(*p) || (*p == '$')) {
					p = parse_int(p, &value, num);
					if ((value >= 0) && (value < 16)) {
						for (i = 0; i < CH_MAX; i++) {
							if (ch & (1 << i)) {
								append_event(score, i, TYPE_ENV_BASE, value, 0, 0);
							}
						}
					} else {
						printf("%d: invalid value in volume envelope.\n", num);
					}
				} else {
					printf("%d: no value in volume envelope.\n", num);
				}
			} else {
				// `
				printf("%d: syntax error. (@%c)\n", num, *p);
				p++;
			}
		} else if (strchr("cdefgab", *p)) {
			// 
			value = tone_inx[*p - 'a'];
			p++;
			if (*p == '+') {
				value++;
				p++;
			} else if (*p == '-') {
				value--;
				p++;
			}

			q = p;
			for (i = 0; i < CH_MAX; i++) {
				if (ch & (1 << i)) {
					p = parse_len(q, &len, score->ch[i].base_len, score->ch[i].def_len, num);
					append_event(score, i, TYPE_NOTE, value + score->ch[i].base_tone, len, score->ch[i].quant);
					score->ch[i].last_note = score->ch[i].last;
				}
			}
		} else if (*p == 'r') {
			// x
			p++;

			q = p;
			for (i = 0; i < CH_MAX; i++) {
				if (ch & (1 << i)) {
					p = parse_len(q, &len, score->ch[i].base_len, score->ch[i].def_len, num);
					append_event(score, i, TYPE_REST, 0, len, score->ch[i].quant);
					score->ch[i].last_note = score->ch[i].last;
				}
			}
		} else if (*p == '!') {
			// Y
			p++;

			q = name;
			while (isalpha(*p) || (*p == '_')) {
				*(q++) = toupper(*p++);
			}
			*q = '\0';

			value = -1;
			for (i = 0; i < score->rpcnt; i++) {
				if (strcmp(name, score->rhythm[i].name) == 0) {
					value = i;
					break;
				}
			}

			if (value >= 0) {
				q = p;
				if (ch & (1 << 3)) {
					p = parse_len(q, &len, score->ch[3].base_len, score->ch[3].def_len, num);
					append_event(score, 3, TYPE_NOTE, value+32+2*12, len, score->ch[3].quant);
					score->ch[3].last_note = score->ch[3].last;
				}
			} else {
				printf("%d: undefined rhythm name. (%s)\n", num, name);
			}
		} else if (*p == '&') {
			// ^C
			p++;

			q = p;
			for (i = 0; i < CH_MAX; i++) {
				if (ch & (1 << i)) {
					p = parse_len(q, &len, score->ch[i].base_len, score->ch[i].def_len, num);
					if (score->ch[i].last_note > 0) {
						append_event(score, i, TYPE_TIE, 0, 0, score->ch[i].quant);
						ev = &score->events[score->ch[i].last_note];
						append_event(score, i, ev->type, ev->value, len, score->ch[i].quant);
					}
				}
			}
		} else if (*p == 'p') {
			// ŉ
			p++;

			for (i = 0; i < CH_MAX; i++) {
				if (ch & (1 << i)) {
					append_event(score, i, TYPE_PERCUSS, 0, 0, 0);
				}
			}
		} else if (*p == '>') {
			// IN^[u㏸
			for (i = 0; i < CH_MAX; i++) {
				if (ch & (1 << i)) {
					score->ch[i].base_tone += 12;
				}
			}
			p++;
		} else if (*p == '<') {
			// IN^[u~
			for (i = 0; i < CH_MAX; i++) {
				if (ch & (1 << i)) {
					score->ch[i].base_tone -= 12;
				}
			}
			p++;
		} else if (*p == 'o') {
			// IN^[uw
			p++;
			if (isdigit(*p) || (*p == '$')) {
				p = parse_int(p, &value, num);
				if ((value >= 2) && (value < 8)) {
					for (i = 0; i < CH_MAX; i++) {
						if (ch & (1 << i)) {
							score->ch[i].base_tone = value * 12;
						}
					}
				} else {
					printf("%d: invalid value in octave.\n", num);
				}
			} else {
				printf("%d: no value in octave.\n", num);
			}
		} else if (*p == 't') {
			// e|w
			p++;
			if (isdigit(*p) || (*p == '$')) {
				p = parse_int(p, &value, num);
				if ((BASE_TICK % value) == 0) {
					for (i = 0; i < CH_MAX; i++) {
						if (ch & (1 << i)) {
							score->ch[i].base_len = BASE_TICK / value;
						}
					}
				} else {
					printf("%d: invalid value in octave.\n", num);
				}
			} else {
				printf("%d: no value in tempo.\n", num);
			}
		} else if (*p == 'l') {
			// ȗ̉w
			p++;
			if (isdigit(*p) || (*p == '$')) {
				p = parse_int(p, &value, num);
				for (i = 0; i < CH_MAX; i++) {
					if (ch & (1 << i)) {
						if ((score->ch[i].base_len % value) == 0) {
							score->ch[i].def_len = value;
						} else {
							printf("%d: invalid value in default length.\n", num);
						}
					}
				}
			} else {
				printf("%d: no value in default length.\n", num);
			}
		} else if (*p == 'q') {
			// NH^CYw
			p++;
			if (isdigit(*p) || (*p == '$')) {
				p = parse_int(p, &value, num);
				if ((value > 0) && (value <= 8)) {
					for (i = 0; i < CH_MAX; i++) {
						if (ch & (1 << i)) {
							score->ch[i].quant = value;
						}
					}
				} else {
					printf("%d: invalid value in quantize.\n", num);
				}
			} else {
				printf("%d: no value in quantize.\n", num);
			}
		} else if (*p == 'm') {
			// Q
			p++;
			if (isdigit(*p) || (*p == '$')) {
				p = parse_int(p, &value, num);
				if ((value > 0) && (value <= 127)) {
					for (i = 0; i < CH_MAX; i++) {
						if (ch & (1 << i)) {
							append_event(score, i, TYPE_CALL, value, 0, 0);
						}
					}
				} else {
					printf("%d: invalid value in mimic.\n", num);
				}
			} else {
				printf("%d: no value in mimic.\n", num);
			}
		} else if (*p == ':') {
			// gbN[v
			p++;
			for (i = 0; i < CH_MAX; i++) {
				if (ch & (1 << i)) {
					append_event(score, i, TYPE_LABEL, 0, 0, 0);
				}
			}
		} else if (*p == '[') {
			// s[g(Jn)
			p++;
			for (i = 0; i < CH_MAX; i++) {
				if (ch & (1 << i)) {
					if (score->ch[i].nest < 2) {
						len  = score->ch[i].label;
						nest = score->ch[i].nest;
						score->ch[i].assigned_label[nest] = len;
						score->ch[i].exist_skip[nest] = 0;
						append_event(score, i, TYPE_LABEL, 0, len, 0);
						score->ch[i].label++;
						score->ch[i].nest++;
					} else {
						printf("%d: repeat nest over.\n", num);
					}
				}
			}
		} else if (*p == '|') {
			// s[g(XLbv)
			p++;
			for (i = 0; i < CH_MAX; i++) {
				if (ch & (1 << i)) {
					nest = score->ch[i].nest - 1;
					len  = score->ch[i].assigned_label[nest];
					append_event(score, i, TYPE_SKIP, 0, len, 0);
					score->ch[i].save_base_tone[nest] = score->ch[i].base_tone;
					score->ch[i].save_base_len[nest]  = score->ch[i].base_len;
					score->ch[i].save_def_len[nest]   = score->ch[i].def_len;
					score->ch[i].save_quant[nest]     = score->ch[i].quant;
					score->ch[i].exist_skip[nest]     = 1;
				}
			}
		} else if (*p == ']') {
			// s[g(I)
			p++;
			if (isdigit(*p) || (*p == '$')) {
				p = parse_int(p, &value, num);
				if ((value > 0) && (value <= 127)) {
					for (i = 0; i < CH_MAX; i++) {
						if (ch & (1 << i)) {
							score->ch[i].nest--;
							nest = score->ch[i].nest;
							len  = score->ch[i].assigned_label[nest];
							append_event(score, i, TYPE_REPEAT, value, len, nest);
							if (score->ch[i].exist_skip[nest] != 0) {
								score->ch[i].base_tone = score->ch[i].save_base_tone[nest];
								score->ch[i].base_len  = score->ch[i].save_base_len[nest];
								score->ch[i].def_len   = score->ch[i].save_def_len[nest];
								score->ch[i].quant     = score->ch[i].save_quant[nest];
							}
						}
					}
				} else {
					printf("%d: invalid value in repeat.\n", num);
				}
			} else {
				printf("%d: no value in repeat.\n", num);
			}
		} else if (*p == '(') {
			// QƔ͈(Jn)
			p++;
			for (i = 0; i < CH_MAX; i++) {
				if (ch & (1 << i)) {
					len  = score->ch[i].label;
					append_event(score, i, TYPE_LABEL, 0, len, 0);
					score->ch[i].ref_ev = score->evcnt;
					score->ch[i].label++;
				}
			}
		} else if (*p == ')') {
			// QƔ͈(I)
			p++;
			if (isdigit(*p) || (*p == '$')) {
				p = parse_int(p, &value, num);
				if ((value >= 1) && (value <= 127)) {
					for (i = 0; i < CH_MAX; i++) {
						if (ch & (1 << i)) {
							if (score->ch[i].ref_ev >= 0) {
								len = score->events[score->ch[i].ref_ev].len;
								append_event(score, i, TYPE_RET, 0, len, 0);
								score->events[score->ch[i].ref_ev].value = value;
								score->ch[i].ref_ev = -1;
							}
						}
					}
				} else {
					printf("%d: invalid value in reference.\n", num);
				}
			} else {
				printf("%d: no value in repeat.\n", num);
			}
		} else if (*p == '/') {
			// gbNI[
			p++;
			for (i = 0; i < CH_MAX; i++) {
				if (ch & (1 << i)) {
					append_event(score, i, TYPE_END, 0, 0, 0);
				}
			}
		} else {
			// `
			printf("%d: syntax error. (%c)\n", num, *p);
			p++;
		}
	}
}

int parse(score_t *score, FILE *fp) {
	char line[1024];
	char *p;
	int ch;
	int def;
	int num;
	int i;

	for (i = 0; i < CH_MAX; i++) {
		score->ch[i].base_tone = 4*12;
		score->ch[i].base_len  = BASE_TICK / 120;
		score->ch[i].def_len   = 4;
		score->ch[i].quant     = 7;
		score->ch[i].label     = 1;
		score->ch[i].nest      = 0;
		score->ch[i].ref_ev    = -1;
	}

	num = 0;
	while (!feof(fp)) {
		p = fgets(line, sizeof(line), fp);
		if (p == NULL) break;
		num++;

		ch  = 0;
		def = 0;
		while (*p) {
			if (isspace(*p)) {
				if (ch) break;
				p++;
			} else if (*p == ';') {
				// Rg
				while (*p) p++;
			} else if (strchr("ABCD", *p)) {
				// `lw
				ch |= 1 << (*p - 'A');
				p++;
			} else if (*p == '#') {
				// f[^`
				p++;
				i = 0;
				while ((i < 3) && isalnum(*p)) {
					def = (def << 8) | *p;
					p++;
					i++;
				}
				if (isspace(*p)) {
					break;
				} else {
					printf("%d: unknown define name.)\n", num);
				}
			} else {
				printf("%d: unknown channel. (%c)\n", num, *p);
				p++;
			}
		}

		if (def != 0) {
			parse_define(score, def, p, num);
		} else {
			parse_score(score, ch, p, num);
		}
	}

	return 0;
}

// note1 ->         label1 -> note2 -> skip1  -> note3 -> repeat1 -> note4
// note1 -> jmp2 -> label1 -> note2 -> label2 -> note3 -> repeat1 -> note4
// note1 -> jmp2 -> label1 -> note3 -> label2 -> note2 -> repeat1 -> note4

int insert_loop_event(score_t *score) {
	int head[LABEL_MAX];
	int skip[LABEL_MAX];
	int tail[LABEL_MAX];
	event_t *ev;
	event_t *end_ev;
	int pos, prev;
	int pos_head, pos_tail;
	int track_loop;
	int i, j, k;

	for (i = 0; i < CH_MAX; i++) {
		memset(head, 0, sizeof(head));
		memset(skip, 0, sizeof(skip));
		memset(tail, 0, sizeof(tail));

		track_loop = 0;
		ev     = NULL;
		end_ev = NULL;
		prev   = 0;
		pos    = score->ch[i].head;
		while (pos > 0) {
			ev = score->events + pos;
			if ((ev->type == TYPE_LABEL) && (ev->len == 0)) {
				track_loop = 1;
			} else if (ev->type == TYPE_END) {
				end_ev = ev;
			}
			if (ev->type == TYPE_LABEL) {
				head[ev->len] = prev;
			} else if (ev->type == TYPE_SKIP) {
				skip[ev->len] = prev;
			} else if (ev->type == TYPE_REPEAT) {
				tail[ev->len] = prev;
			} else if (ev->type == TYPE_RET) {
				tail[ev->len] = prev;
			}
			prev = pos;
			pos  = ev->next;
		}
		if (track_loop) {
			if (end_ev != NULL) {
				end_ev->type = TYPE_JMP;
			} else {
				append_event(score, i, TYPE_JMP, 0, 0, 0);
			}
		} else {
			if ((ev != NULL) && (ev->type != TYPE_JMP) && (ev->type != TYPE_RET) && (ev->type != TYPE_END)) {
				append_event(score, i, TYPE_END, 0, 0, 0);
			}
		}

		// nest̐[珇ɕϊ
		for (j = 1; j >= 0; --j) {
			for (k = 1; tail[k] > 0; k++) {
				if (skip[k] > 0) {
					ev = score->events + tail[k];
					pos = ev->next;
					ev = score->events + pos;
					if ((ev->type == TYPE_REPEAT) && (ev->quant == j)) {
						// WvCxgǉ
						insert_event(score, i, head[k], TYPE_JMP, 0, score->ch[i].label, 0);
						// XLbvCxgxɕύX
						ev = score->events + skip[k];
						pos = ev->next;
						ev = score->events + pos;
						ev->type = TYPE_LABEL;
						ev->len  = score->ch[i].label;
						score->ch[i].label++;
						// s[g擪猳XLbv܂łƌXLbv烊s[g܂łւ
						ev = score->events + head[k];
						ev = score->events + ev->next;
						ev = score->events + ev->next;
						pos_head = ev->next;
						ev->next = score->events[pos].next;
						ev = score->events + tail[k];
						pos_tail = ev->next;
						ev->next = pos;
						score->events[pos].next = pos_head;
						ev = score->events + skip[k];
						ev->next = pos_tail;
					}
				}
			}
		}
	}

	return 0;
}

int insert_label_event(score_t *score) {
	call_t tbl[LABEL_MAX];
	event_t *ev;
	int pos, ref;
	int i;

	for (i = 0; i < CH_MAX; i++) {
		memset(tbl, 0, sizeof(tbl));

		pos = score->ch[i].head;
		while (pos > 0) {
			ev = score->events + pos;
			if ((ev->type == TYPE_LABEL) && (ev->value > 0)) {
				tbl[ev->value].pos   = pos;
				tbl[ev->value].count = 0;
			}
			pos = ev->next;
		}

		pos = score->ch[i].head;
		while (pos > 0) {
			ev = score->events + pos;
			if (ev->type == TYPE_CALL) {
				ref = tbl[ev->value].pos;
				if (ref > 0) {
					if (tbl[ev->value].count > 0) {
						ev->len = score->ch[i].label;
						insert_event(score, i, ref, TYPE_LABEL, ev->value, ev->len, 0);
						score->ch[i].label++;
					} else {
						ev->len = score->events[ref].len;
					}
					tbl[ev->value].count++;
				}
			}
			pos = ev->next;
		}
	}
}

int insert_tone_event(score_t *score) {
	event_t *ev;
	int pos, prev, sect;
	int low, high;
	int last_low, last_high;
	int i;

	for (i = 0; i < CH_MAX; i++) {
		sect = 0;
		prev = 0;
		pos  = score->ch[i].head;
		low  = 0;
		high = 0;
		last_low  = 0;
		last_high = 0;
		while (pos > 0) {
			ev = score->events + pos;
			if (ev->type == TYPE_NOTE) {
				if (low == 0) {
					sect = prev;
					low  = ev->value;
					high = ev->value;
				} else {
					if (ev->value < low) {
						if (ev->value <= high - TONE_RANGE) {
							insert_event(score, i, sect, TYPE_TONE_BASE, low, high, 0);
							sect = prev;
							low  = ev->value;
							high = ev->value;
						} else {
							low  = ev->value;
						}
					} else if (ev->value > high) {
						if (ev->value >= low + TONE_RANGE) {
							insert_event(score, i, sect, TYPE_TONE_BASE, low, high, 0);
							sect = prev;
							low  = ev->value;
							high = ev->value;
						} else {
							high = ev->value;
						}
					}
				}
			} else if (ev->type == TYPE_LABEL) {
				if (low != 0) {
					insert_event(score, i, sect, TYPE_TONE_BASE, low, high, 0);
					sect = prev;
					last_low  = low;
					last_high = high;
					low  = 0;
					high = 0;
				}
			} else if ((ev->type == TYPE_JMP) || (ev->type == TYPE_CALL)) {
				if (low == 0) {
					insert_event(score, i, prev, TYPE_TONE_BASE, last_low, last_high, 0);
					sect = pos;
				}
			}
			prev = pos;
			pos  = ev->next;
		}
		if (low > 0) {
			insert_event(score, i, sect, TYPE_TONE_BASE, low, high, 0);
		}
	}

	return 0;
}

int merge_len_table(lentbl_t *tbl1, lentbl_t *tbl2) {
	unsigned long value1, value2;
	int mask, cnt;
	int i, j;
	int r = 0;

	mask = 0;
	cnt  = 0;
	for (i = 0; i < tbl1->count; i++) {
		for (j = 0; j < tbl2->count; j++) {
			if ((tbl1->len[i]   == tbl2->len[j]) &&
			    (tbl1->quant[i] == tbl2->quant[j])) {
				mask |= (1 << j);
				cnt++;
			}
		}
	}
	if (cnt == tbl1->count) {
		tbl1->count = 0;
		r = 1;
	} else if (cnt == tbl2->count) {
		tbl2->count = 0;
		r = 2;
	} else if ((tbl2->count - cnt) <= (LENTBL_SIZE - tbl1->count)) {
		for (j = 0; j < tbl2->count; j++) {
			if ((mask & (1 << j)) == 0) {
				value2 = (tbl2->len[j] << 8) | tbl2->quant[j];
				for (i = tbl1->count; i > 0; --i) {
					value1 = (tbl1->len[i-1] << 8) | tbl1->quant[i-1];
					if (value1 <= value2) break;
					tbl1->len[i]   = tbl1->len[i-1];
					tbl1->quant[i] = tbl1->quant[i-1];
				}
				tbl1->len[i]   = tbl2->len[j];
				tbl1->quant[i] = tbl2->quant[j];
				tbl1->count++;
			}
		}
		tbl2->count = 0;
		r = 2;
	}

	return r;
}

int insert_lentbl_event(score_t *score) {
	lentbl_t *tbl1;
	lentbl_t *tbl2;
	event_t *ev;
	unsigned long value1, value2;
	int pos, prev, sect;
	int i, j, k;

	// e[u̍쐬
	for (i = 0; i < CH_MAX; i++) {
		sect = 0;
		prev = 0;
		pos  = score->ch[i].head;
		tbl1 = NULL;
		while (pos > 0) {
			ev = score->events + pos;
			if ((ev->type == TYPE_NOTE) || (ev->type == TYPE_REST)) {
				if (tbl1 == NULL) {
					sect = prev;
					tbl1 = score->lentbl + score->ltcnt;
					score->ltcnt++;
					tbl1->len[0]   = ev->len;
					tbl1->quant[0] = ev->quant;
					tbl1->count = 1;
				} else {
					for (j = 0; j < tbl1->count; j++) {
						if ((tbl1->len[j]   == ev->len  ) &&
						    (tbl1->quant[j] == ev->quant)) break;
					}
					if (j >= tbl1->count) {
						if (tbl1->count >= LENTBL_SIZE) {
							insert_event(score, i, sect, TYPE_LEN_TABLE, score->ltcnt-1, 0, 0);
							sect = prev;

							tbl1 = score->lentbl + score->ltcnt;
							score->ltcnt++;
							tbl1->len[0]   = ev->len;
							tbl1->quant[0] = ev->quant;
							tbl1->count = 1;
						} else {
							value2 = (ev->len << 8) | ev->quant;
							for (j = tbl1->count; j > 0; --j) {
								value1 = (tbl1->len[j-1] << 8) | tbl1->quant[j-1];
								if (value1 <= value2) break;
								tbl1->len[j]   = tbl1->len[j-1];
								tbl1->quant[j] = tbl1->quant[j-1];
							}
							tbl1->len[j]   = ev->len;
							tbl1->quant[j] = ev->quant;
							tbl1->count++;
						}
					}
				}
			} else if (ev->type == TYPE_LABEL) {
				if (tbl1 != NULL) {
					insert_event(score, i, sect, TYPE_LEN_TABLE, score->ltcnt-1, 0, 0);
					sect = prev;
					tbl1 = NULL;
				}
			} else if ((ev->type == TYPE_JMP) || (ev->type == TYPE_CALL)) {
				if (tbl1 == NULL) {
					insert_event(score, i, prev, TYPE_LEN_TABLE, score->ltcnt-1, 0, 0);
					sect = pos;
				}
			}
			prev = pos;
			pos  = ev->next;
		}
		if (tbl1) {
			insert_event(score, i, sect, TYPE_LEN_TABLE, score->ltcnt-1, 0, 0);
		}
	}

	// e[u̓
	for (i = 0; i < score->ltcnt; i++) {
		tbl1 = score->lentbl + i;
		if (tbl1->count > 0) {
			for (j = 0; j < score->ltcnt; j++) {
				tbl2 = score->lentbl + j;
				if ((i != j) && (tbl2->count > 0)) {
					switch (merge_len_table(tbl1, tbl2)) {
					case 1:
						for (k = 0; k < CH_MAX; k++) {
							pos = score->ch[k].head;
							while (pos > 0) {
								ev = score->events + pos;
								if ((ev->type == TYPE_LEN_TABLE) && (ev->value == i)) {
									ev->value = j;
								}
								pos = ev->next;
							}
						}
						break;
					case 2:
						for (k = 0; k < CH_MAX; k++) {
							pos = score->ch[k].head;
							while (pos > 0) {
								ev = score->events + pos;
								if ((ev->type == TYPE_LEN_TABLE) && (ev->value == j)) {
									ev->value = i;
								}
								pos = ev->next;
							}
						}
						break;
					}
				}
			}
		}
	}

	// e[ũx[Xʒuݒ
	pos = (score->rpcnt+1)/2;
	for (i = 0; i < score->ltcnt; i++) {
		tbl1 = score->lentbl + i;
		tbl1->base = pos;
		pos += tbl1->count;
	}

	return 0;
}

int optimize_event(score_t *score) {
	opttbl_t tbl[LABEL_MAX];
	event_t *ev;
	int pos, prev;
	int label, label2;
	int low, high, ltbl;
	int i, b, f;

	for (i = 0; i < CH_MAX; i++) {
		for (;;) {
			memset(tbl, 0, sizeof(tbl));

			low   = 0;
			high  = 0;
			ltbl  = -1;
			label = -1;
			prev  = -2;
			pos = score->ch[i].head;
			while (pos > 0) {
				ev = score->events + pos;
				if (ev->type == TYPE_LABEL) {
					if (label >= 0) {
						tbl[ev->len].link = label;
					} else {
						tbl[ev->len].link = -1;
					}
					label = ev->len;
					tbl[label].near.pos  = prev;
					tbl[label].near.low  = low;
					tbl[label].near.high = high;
					tbl[label].near.ltbl = ltbl;
				} else if (ev->type == TYPE_REPEAT) {
					tbl[ev->len].far.pos  = prev;
					tbl[ev->len].far.low  = low;
					tbl[ev->len].far.high = high;
					tbl[ev->len].far.ltbl = ltbl;
				} else if (ev->type == TYPE_JMP) {
					tbl[ev->len].far.pos  = prev;
					tbl[ev->len].far.low  = low;
					tbl[ev->len].far.high = high;
					tbl[ev->len].far.ltbl = ltbl;
					low   = 0;
					high  = 0;
					ltbl  = -1;
					label = -1;
					prev  = -1;
				} else if (ev->type == TYPE_CALL) {
					tbl[ev->len].far.pos  = prev;
					tbl[ev->len].far.low  = low;
					tbl[ev->len].far.high = high;
					tbl[ev->len].far.ltbl = ltbl;
				} else if (ev->type == TYPE_TONE_BASE) {
					prev = pos;
					low  = ev->value;
					high = ev->len;
				} else if (ev->type == TYPE_LEN_TABLE) {
					ltbl = ev->value;
				} else if (ev->type == TYPE_NOTE) {
					label = -1;
				}
				pos = ev->next;
			}

			ltbl  = -1;
			label = -1;
			pos = score->ch[i].head;
			while (pos > 0) {
				ev = score->events + pos;
				if (ev->type == TYPE_LABEL) {
					label = ev->len;
				} else if (ev->type == TYPE_JMP) {
					ltbl  = -1;
					label = -1;
				} else if (ev->type == TYPE_TONE_BASE) {
/*
					if (label >= 0) {
						printf("label:%3d (%3d: %2d %2d) (%3d: %2d %2d) (%3d: %2d %2d)\n",
							label, 
							pos, ev->value, ev->len,
							tbl[label].near.pos, tbl[label].near.low, tbl[label].near.high,
							tbl[label].far.pos, tbl[label].far.low, tbl[label].far.high);
						label2 = tbl[label].link;
						while (label2 >= 0) {
							printf("          label:%3d    (%3d: %2d %2d) (%3d: %2d %2d)\n",
								label2, 
								tbl[label2].near.pos, tbl[label2].near.low, tbl[label2].near.high,
								tbl[label2].far.pos, tbl[label2].far.low, tbl[label2].far.high);
							label2 = tbl[label2].link;
						}
					}
*/
					if (label >= 0) {
						low  = ev->value;
						high = ev->len;
						label2 = label;
						while (label2 >= 0) {
							if ((tbl[label2].near.pos >= 0) && (tbl[label2].link < 0)) {
								if (tbl[label2].near.low  < low ) low  = tbl[label2].near.low;
								if (tbl[label2].near.high > high) high = tbl[label2].near.high;
							}
							if (tbl[label2].far.pos >= 0) {
								if (tbl[label2].far.low  < low ) low  = tbl[label2].far.low;
								if (tbl[label2].far.high > high) high = tbl[label2].far.high;
							}
							label2 = tbl[label2].link;
						}
						if ((high - low + 1) <= TONE_RANGE) {
							b = 0;
							f = 1;
							b |= (ev->value != low)  ? (f<<2) : 0;
							b |= (ev->len   != high) ? (f<<3) : 0;
							f <<= 4;
							ev->value = low;
							ev->len   = high;
							label2 = label;
							while (label2 >= 0) {
								if (tbl[label2].near.pos == -2) b |= 1;
								if ((tbl[label2].near.pos >= 0) && (tbl[label2].link < 0)) {
									b |= (score->events[tbl[label2].near.pos].value != low)  ? (f   ) : 0;
									b |= (score->events[tbl[label2].near.pos].len   != high) ? (f<<1) : 0;
									score->events[tbl[label2].near.pos].value = low;
									score->events[tbl[label2].near.pos].len   = high;
								}
								if (tbl[label2].far.pos == -2) b |= 1;
								if (tbl[label2].far.pos >= 0) {
									b |= (score->events[tbl[label2].far.pos].value != low)  ? (f<<2) : 0;
									b |= (score->events[tbl[label2].far.pos].len   != high) ? (f<<3) : 0;
									score->events[tbl[label2].far.pos].value = low;
									score->events[tbl[label2].far.pos].len   = high;
								}
								f <<= 4;
								label2 = tbl[label2].link;
							}
							ev->quant = (b & 1) ? 0 : 1;
							if (b & -4) {
//								printf("----(%02x)\n", b);
								break;
							}
						} else {
							ev->quant = 0;
						}
					}
					low  = ev->value;
					high = ev->len;
				} else if (ev->type == TYPE_LEN_TABLE) {
/*
					if (label >= 0) {
						printf("label:%3d ltbl:%3d ltbl:%3d ltbl:%3d\n",
							label, ev->value, ltbl, tbl[label].ltbl);
					}
*/
					if ((label >= 0) && (ev->quant == 0)) {
						f = 1;
						b = (ev->value != ltbl) ? f : 0;
						f <<= 1;
						label2 = label;
						while (label2 >= 0) {
							b |= (ev->value != tbl[label2].near.ltbl) ? (f   ) : 0;
							b |= (ev->value != tbl[label2].far.ltbl)  ? (f<<1) : 0;
							f <<= 2;
							label2 = tbl[label2].link;
						}
						ev->quant = (b) ? 0 : 1;
					}
					ltbl = ev->value;
				} else if (ev->type == TYPE_NOTE) {
					label = -1;
				}
				pos = ev->next;
			}

			if (pos == 0) break;
		}
	}

	return 0;
}

int assign_label_offset(score_t *score) {
	int locate[LABEL_MAX];
	event_t *ev;
	int pos, cnt;
	int i, j, n;

	for (i = 0; i < CH_MAX; i++) {
		memset(locate, 0, sizeof(locate));

		for (j = 0; j < 3; j++) {
			cnt = 0;
			pos  = score->ch[i].head;
			while (pos > 0) {
				ev = score->events + pos;
				if (ev->type == TYPE_NOTE) {
					cnt++;
				} else if (ev->type == TYPE_REST) {
					cnt++;
				} else if (ev->type == TYPE_REG1) {
					cnt += 2;
				} else if (ev->type == TYPE_REG2) {
					cnt += 2;
				} else if ((ev->type == TYPE_TONE_BASE) && (ev->quant == 0)) {
					cnt += 2;
				} else if ((ev->type == TYPE_LEN_TABLE) && (ev->quant == 0)) {
					cnt += 2;
				} else if (ev->type == TYPE_ENV_BASE) {
					cnt += 2;
				} else if (ev->type == TYPE_TIE) {
					cnt++;
				} else if (ev->type == TYPE_PERCUSS) {
					cnt++;
				} else if (ev->type == TYPE_REPEAT) {
					if (j == 0) {
						cnt += 3;
					} else if (j == 1) {
						n = locate[ev->len] - (cnt+2);
						if ((ev->value <= 5) && (n > -32)) {
							cnt += 2;
						} else {
							cnt += 3;
						}
					} else {
						n = locate[ev->len] - (cnt+2);
						if ((ev->value <= 5) && (n > -32)) {
							cnt += 2;
							ev->quant &= ~2;
							ev->len = (locate[ev->len] - cnt) & 0x1f;
						} else {
							cnt += 3;
							ev->quant |= 2;
							if (cnt <= locate[ev->len]+256) {
								ev->len = (locate[ev->len] - cnt) & 0xff;
							} else {
								printf("repeat distance error.\n");
							}
						}
					}
				} else if (ev->type == TYPE_JMP) {
					if (j == 0) {
						cnt += 3;
					} else if (j == 1) {
						n = locate[ev->len] - (cnt+2);
						if ((n > -64) && (n < 64)) {
							cnt += 2;
						} else {
							cnt += 3;
						}
					} else {
						n = locate[ev->len] - (cnt+2);
						if ((n > -64) && (n < 64)) {
							ev->quant = 0;
							cnt += 2;
						} else {
							ev->quant = 1;
							cnt += 3;
						}
						ev->len = (locate[ev->len] - cnt) & 0xffff;
					}
				} else if (ev->type == TYPE_CALL) {
					cnt += 3;
					if (j >= 2) {
						ev->len = (locate[ev->len] - cnt) & 0xffff;
					}
				} else if (ev->type == TYPE_RET) {
					cnt += 1;
				} else if (ev->type == TYPE_END) {
					cnt += 1;
				} else if (ev->type == TYPE_LABEL) {
					if (j < 2) {
						locate[ev->len] = cnt;
					}
				}
				pos  = ev->next;
			}
		}
	}

	return 0;
}

int assign_envtable_base(score_t *score) {
	int count;
	int i, n;

	count = 2 + score->dt.count*2;
	for (i = 0; i < ENV_MAX; i++) {
		score->env[i].base = count;
		count += score->env[i].count;
	}

	for (i = 0; i < score->rpcnt; i++) {
		n = score->rhythm[i].data & 15;
		if (n > 0) {
			n = score->env[n-1].base / 2;
			if (n <= 15) {
				score->rhythm[i].data &= 0xfff0;
				score->rhythm[i].data |= n;
			} else {
				printf("rhythm envelope error. (%s)\n", score->rhythm[i].name);
			}
		}
	}

	return 0;
}

int write_cmd(score_t *score, FILE *fp, const char *prefix) {
	int valid;
	int i;

	valid = 0;
	for (i = 0; i < CH_MAX; i++) {
		if (score->ch[i].head > 0) {
			valid |= (1 << i);
		}
	}

	fprintf(fp, "%s_PlayData:\n", prefix);
	fprintf(fp, "\t.dw\t$%04x, %s_LenTable, %s_EnvTable\n", valid, prefix, prefix);
	fprintf(fp, "\t.dw\t", prefix);
	for (i = 0; i < CH_MAX; i++) {
		if (valid & (1 << i)) {
			fprintf(fp, "%s_Seq%c", prefix, i+'A');
			if (((valid & 15) & ~((1<<(i+1))-1)) != 0) {
				fprintf(fp, ", ");
			}
		}
	}
	fprintf(fp, "\n");

	return 0;
}

int write_data(score_t *score, FILE *fp, const char *prefix) {
	unsigned char buf[4];
	event_t *ev;
	lentbl_t *tbl;
	int pos, cnt;
	int tone_base;
	int i, j, n;

	// e[u
	fprintf(fp, "%s_LenTable:\n", prefix);
	cnt = 0;
	for (i = 0; i < score->rpcnt; i++) {
		if ((cnt % 4) == 0) {
			fprintf(fp, "\t.dw\t");
		} else {
			fprintf(fp, ", ");
		}
		fprintf(fp, "$%04x", score->rhythm[i].data);
		if ((cnt % 4) == 3) {
			fprintf(fp, "\n");
		}
		cnt++;
	}
	if ((score->rpcnt % 2) != 0) {
		fprintf(fp, ", $0000");
	}
	if ((cnt % 4) != 0) {
		fprintf(fp, "\n");
	}
	cnt = 0;
	for (i = 0; i < score->ltcnt; i++) {
		tbl = score->lentbl + i;
		for (j = 0; j < tbl->count; j++) {
			if ((cnt % 4) == 0) {
				fprintf(fp, "\t.dw\t");
			} else {
				fprintf(fp, ", ");
			}
			if ((tbl->quant[j] == 8) || (tbl->len[j] <= 4)) {
				n = 0;
			} else {
				n = tbl->len[j] * (8 - tbl->quant[j]) / 8;
				if (n < 4) n = 4;
				if (tbl->len[j] - n < 4) n = tbl->len[j] - 4;
			}
			fprintf(fp, "%4d, %4d", tbl->len[j]-n, n);
			if ((cnt % 4) == 3) {
				fprintf(fp, "\n");
			}
			cnt++;
		}
	}
	if ((cnt % 4) != 0) {
		fprintf(fp, "\n");
	}

	// Gx[ve[u
	cnt = 2;
	fprintf(fp, "%s_EnvTable:\n", prefix);
	fprintf(fp, "\t.db\t$00, $00");
	for (j = 0; j < score->dt.count; j++) {
		if ((cnt % 8) == 0) {
			fprintf(fp, "\t.db\t");
		} else {
			fprintf(fp, ", ");
		}
		fprintf(fp, "$%02x", score->dt.data[j] & 0xff);
		if ((cnt % 8) == 7) {
			fprintf(fp, "\n");
		}
		cnt++;

		if ((cnt % 8) == 0) {
			fprintf(fp, "\t.db\t");
		} else {
			fprintf(fp, ", ");
		}
		fprintf(fp, "$%02x", score->dt.data[j] >> 8);
		if ((cnt % 8) == 7) {
			fprintf(fp, "\n");
		}
		cnt++;
	}
	for (i = 0; i < ENV_MAX; i++) {
		for (j = 0; j < score->env[i].count; j++) {
			if ((cnt % 8) == 0) {
				fprintf(fp, "\t.db\t");
			} else {
				fprintf(fp, ", ");
			}
			fprintf(fp, "$%02x", score->env[i].data[j]);
			if ((cnt % 8) == 7) {
				fprintf(fp, "\n");
			}
			cnt++;
		}
	}
	if ((cnt % 8) != 0) {
		fprintf(fp, "\n");
	}

	// V[PXf[^
	for (i = 0; i < CH_MAX; i++) {
		if (score->ch[i].head > 0) {
			fprintf(fp, "%s_Seq%c:\n", prefix, i+'A');
			pos = score->ch[i].head;
			cnt = 0;
			tone_base = 0;
			tbl = NULL;
			while (pos > 0) {
				ev = score->events + pos;
				switch (ev->type) {
				case TYPE_NOTE:
					if (tbl) {
						for (j = 0; j < LENTBL_SIZE; j++) {
							if (tbl->len[j] == ev->len) break;
						}
						if (j >= LENTBL_SIZE) j = 0;
					} else {
						j = 0;
					}
					buf[0] = (j << 4) | (ev->value - tone_base + 1);
					n = 1;
					break;
				case TYPE_REST:
					if (tbl) {
						for (j = 0; j < LENTBL_SIZE; j++) {
							if (tbl->len[j] == ev->len) break;
						}
						if (j >= LENTBL_SIZE) j = 0;
					} else {
						j = 0;
					}
					buf[0] = (j << 4);
					n = 1;
					break;
				case TYPE_REG1:
					buf[0] = 0xf0;
					buf[1] = ev->value;
					n = 2;
					break;
				case TYPE_REG2:
					buf[0] = 0xf1;
					buf[1] = ev->value;
					n = 2;
					break;
				case TYPE_TONE_BASE:
					if (ev->quant == 0) {
						buf[0] = 0xf2;
						buf[1] = ev->value - 2*12;
						n = 2;
					} else {
						n = 0;
					}
					tone_base = ev->value;
					break;
				case TYPE_LEN_TABLE:
					if (ev->quant == 0) {
						buf[0] = 0xf3;
						buf[1] = score->lentbl[ev->value].base;
						n = 2;
					} else {
						n = 0;
					}
					tbl = score->lentbl + ev->value;
					break;
				case TYPE_ENV_BASE:
					buf[0] = 0xf4;
					buf[1] = (ev->value > 0) ? score->env[ev->value-1].base : 0;
					n = 2;
					break;
				case TYPE_TIE:
					buf[0] = 0xf5;
					n = 1;
					break;
				case TYPE_PERCUSS:
					buf[0] = 0xf8;
					n = 1;
					break;
				case TYPE_CALL:
					buf[0] = 0xfa;
					buf[1] = ((ev->len   )&0xff);
					buf[2] = ((ev->len>>8)&0xff);
					n = 3;
					break;
				case TYPE_RET:
					buf[0] = 0xfb;
					n = 1;
					break;
				case TYPE_REPEAT:
					if ((ev->quant & 2) == 0) {
						buf[0] = 0xfc;
						buf[1] = ((ev->len<<3)|((ev->value-2)<<1)|(ev->quant&1)) & 0xff;
						n = 2;
					} else {
						buf[0] = 0xfd;
						buf[1] = ((ev->value<<1)|(ev->quant&1)) & 0xff;
						buf[2] = ev->len & 0xff;
						n = 3;
					}
					break;
				case TYPE_JMP:
					if (ev->quant == 0) {
						buf[0] = 0xfe;
						buf[1] = ev->len & 0x7f;
						n = 2;
					} else {
						buf[0] = 0xfe;
						buf[1] = ((ev->len   )&0x7f)|0x80;
						buf[2] = ((ev->len>>7)&0xff);
						n = 3;
					}
					break;
				case TYPE_END:
					buf[0] = 0xff;
					n = 1;
					break;
				default:
					n = 0;
				}
				for (j = 0; j < n; j++) {
					if ((cnt % 8) == 0) {
						fprintf(fp, "\t.db\t");
					} else {
						fprintf(fp, ", ");
					}
					fprintf(fp, "$%02x", buf[j]);
					if ((cnt % 8) == 7) {
						fprintf(fp, "\n");
					}
					cnt++;
				}
				pos = ev->next;
			}
			if ((cnt % 8) != 0) {
				fprintf(fp, "\n");
			}
		}
	}

	return 0;
}

void dump_event(event_t *events, int pos) {
	static const char *note[] = { "c", "c+", "d", "d+", "e", "f", "f+", "g", "g+", "a", "a+", "b" };
	event_t *ev;
	int tone;
	int i;

	while (pos > 0) {
		ev = events + pos;
		printf("type:%2d value:%5d len:%5d quant:%2d", ev->type, ev->value, ev->len, ev->quant);
		if (ev->type == TYPE_NOTE) {
			tone = ev->value;
			printf(" (o%d%s)", tone/12, note[tone%12]);
		}
		printf("\n");
		pos = ev->next;
	}
}

void dump_lentbl(lentbl_t *lentbl) {
	int i;

	printf("[ ");
	for (i = 0; i < lentbl->count; i++) {
		printf("%d:%d ", lentbl->len[i], lentbl->quant[i]);
	}
	printf("]\n");
}

void dump(score_t *score) {
	int i;

	for (i = 0; i < CH_MAX; i++) {
		printf("==== ch%c ====\n", i+'A');
		dump_event(score->events, score->ch[i].head);
	}
	for (i = 0; i < score->ltcnt; i++) {
		printf("==== lentbl%d ====\n", i);
		dump_lentbl(score->lentbl + i);
	}
}

